@@ -2,7 +2,7 @@
  * PL/R - PostgreSQL support for R as a
  *	      procedural language (PL)
  *
- * Copyright (c) 2003-2009 by Joseph E. Conway
+ * Copyright (c) 2003-2010 by Joseph E. Conway
  * ALL RIGHTS RESERVED
  * 
  * Joe Conway <mail@joeconway.com>
@@ -33,9 +33,6 @@
  *                   to pg types
  */
 #include "plr.h"
-#include "funcapi.h"
-#include "miscadmin.h"
-#include "catalog/catversion.h"
 
 static void pg_get_one_r(char *value, Oid arg_out_fn_oid, SEXP *obj,
 																int elnum);
@@ -44,6 +41,7 @@ static Datum get_trigger_tuple(SEXP rval, plr_function *function,
 									FunctionCallInfo fcinfo, bool *isnull);
 static Datum get_tuplestore(SEXP rval, plr_function *function,
 									FunctionCallInfo fcinfo, bool *isnull);
+static Datum get_simple_array_datum(SEXP rval, Oid typelem, bool *isnull);
 static Datum get_array_datum(SEXP rval, plr_function *function, int col, bool *isnull);
 static Datum get_frame_array_datum(SEXP rval, plr_function *function, int col,
 																bool *isnull);
@@ -96,11 +94,13 @@ pg_scalar_get_r(Datum dvalue, Oid arg_typid, FmgrInfo arg_out_func)
 	{
 		SEXP 	s, t, obj;
 		int		status;
+		Datum	dt_dvalue =  PointerGetDatum(PG_DETOAST_DATUM(dvalue));
+		int		bsize = VARSIZE((bytea *) dt_dvalue);
 
-		PROTECT(obj = get_r_vector(arg_typid, VARSIZE((bytea *) dvalue)));
+		PROTECT(obj = get_r_vector(arg_typid, bsize));
 		memcpy((char *) RAW(obj),
-			   VARDATA((bytea *) dvalue),
-			   VARSIZE((bytea *) dvalue));
+			   VARDATA((bytea *) dt_dvalue),
+			   bsize);
 
 		/*
 		 * Need to construct a call to
@@ -108,7 +108,8 @@ pg_scalar_get_r(Datum dvalue, Oid arg_typid, FmgrInfo arg_out_func)
 		 */
 		PROTECT(t = s = allocList(2));
 		SET_TYPEOF(s, LANGSXP);
-		SETCAR(t, install("unserialize")); t = CDR(t);
+		SETCAR(t, install("unserialize"));
+		t = CDR(t);
 		SETCAR(t, obj);
 
 		PROTECT(result = R_tryEval(s, R_GlobalEnv, &status));
@@ -126,7 +127,7 @@ pg_scalar_get_r(Datum dvalue, Oid arg_typid, FmgrInfo arg_out_func)
 						 errdetail("R expression evaluation error caught in \"unserialize\".")));
 		}
 
-		UNPROTECT(2);
+		UNPROTECT(3);
 	}
 
 	return result;
@@ -144,8 +145,7 @@ pg_array_get_r(Datum dvalue, FmgrInfo out_func, int typlen, bool typbyval, char
 	 * Use the converted values to build an R vector.
 	 */
 	SEXP		result;
-	char	   *value;
-	ArrayType  *v = DatumGetArrayTypeP(dvalue);
+	ArrayType  *v;
 	Oid			element_type;
 	int			i, j, k,
 				nitems,
@@ -157,101 +157,300 @@ pg_array_get_r(Datum dvalue, FmgrInfo out_func, int typlen, bool typbyval, char
 	int			elem_idx = 0;
 	Datum	   *elem_values;
 	bool	   *elem_nulls;
+	bool		fast_track_type;
 
 	/* short-circuit for NULL datums */
 	if (dvalue == (Datum) NULL)
 		return R_NilValue;
 
+	v = DatumGetArrayTypeP(dvalue);
 	ndim = ARR_NDIM(v);
 	element_type = ARR_ELEMTYPE(v);
 	dim = ARR_DIMS(v);
+	nitems = ArrayGetNItems(ARR_NDIM(v), ARR_DIMS(v));
 
-	deconstruct_array(v, element_type,
-					  typlen, typbyval, typalign,
-					  &elem_values, &elem_nulls, &nitems);
-
-	/* array is empty */
-	if (nitems == 0)
+	switch (element_type)
 	{
+		case INT4OID:
+		case FLOAT8OID:
+			fast_track_type = true;
+			break;
+		default:
+			fast_track_type = false;
+	}
+
+	/*
+	 * Special case for pass-by-value data types, if the following conditions are met:
+	 * 		designated fast_track_type
+	 * 		no NULL elements
+	 * 		1 dimensional array only
+	 * 		at least one element
+	 */
+	if (fast_track_type &&
+		 typbyval &&
+		 !ARR_HASNULL(v) &&
+		 (ndim == 1) &&
+		 (nitems > 0))
+	{
+		char	   *p = ARR_DATA_PTR(v);
+
+		/* get new vector of the appropriate type and length */
 		PROTECT(result = get_r_vector(element_type, nitems));
-		UNPROTECT(1);
 
-		return result;
-	}
+		/* keep this in sync with switch above -- fast_track_type only */
+		switch (element_type)
+		{
+			case INT4OID:
+				Assert(sizeof(int) == 4);
+				memcpy(INTEGER_DATA(result), p, nitems * sizeof(int));
+				break;
+			case FLOAT8OID:
+				Assert(sizeof(double) == 8);
+				memcpy(NUMERIC_DATA(result), p, nitems * sizeof(double));
+				break;
+			default:
+				/* Everything else is error */
+				ereport(ERROR,
+						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+						 errmsg("direct array passthrough attempted for unsupported type")));
+		}
 
-	if (ndim == 1)
-		nr = nitems;
-	else if (ndim == 2)
-	{
-		nr = dim[0];
-		nc = dim[1];
-	}
-	else if (ndim == 3)
-	{
-		nr = dim[0];
-		nc = dim[1];
-		nz = dim[2];
+		if (ndim > 1)
+		{
+			SEXP	matrix_dims;
+
+			/* attach dimensions */
+			PROTECT(matrix_dims = allocVector(INTSXP, ndim));
+			for (i = 0; i < ndim; i++)
+				INTEGER_DATA(matrix_dims)[i] = dim[i];
+
+			setAttrib(result, R_DimSymbol, matrix_dims);
+			UNPROTECT(1);
+		}
+
+		UNPROTECT(1);	/* result */
 	}
 	else
-		ereport(ERROR,
-				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-				 errmsg("greater than 3-dimensional arrays are not yet supported")));
-
-	/* get new vector of the appropriate type and length */
-	PROTECT(result = get_r_vector(element_type, nitems));
-
-	/* Convert all values to their R form and build the vector */
-	for (i = 0; i < nr; i++)
 	{
-		for (j = 0; j < nc; j++)
+		deconstruct_array(v, element_type,
+						  typlen, typbyval, typalign,
+						  &elem_values, &elem_nulls, &nitems);
+
+		/* array is empty */
+		if (nitems == 0)
 		{
-			for (k = 0; k < nz; k++)
+			PROTECT(result = get_r_vector(element_type, nitems));
+			UNPROTECT(1);
+
+			return result;
+		}
+
+		if (ndim == 1)
+			nr = nitems;
+		else if (ndim == 2)
+		{
+			nr = dim[0];
+			nc = dim[1];
+		}
+		else if (ndim == 3)
+		{
+			nr = dim[0];
+			nc = dim[1];
+			nz = dim[2];
+		}
+		else
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+					 errmsg("greater than 3-dimensional arrays are not yet supported")));
+
+		/* get new vector of the appropriate type and length */
+		PROTECT(result = get_r_vector(element_type, nitems));
+
+		/* Convert all values to their R form and build the vector */
+		for (i = 0; i < nr; i++)
+		{
+			for (j = 0; j < nc; j++)
 			{
-				Datum		itemvalue;
-				bool		isnull;
-				int			idx = (k * nr * nc) + (j * nr) + i;
-
-				isnull = elem_nulls[elem_idx];
-				itemvalue = elem_values[elem_idx++];
-
-				if (!isnull)
+				for (k = 0; k < nz; k++)
 				{
-					value = DatumGetCString(FunctionCall3(&out_func,
-														  itemvalue,
-														  (Datum) 0,
-														  Int32GetDatum(-1)));
-				}
-				else
-					value = NULL;
+					char	   *value;
+					Datum		itemvalue;
+					bool		isnull;
+					int			idx = (k * nr * nc) + (j * nr) + i;
 
-				/*
-				 * Note that pg_get_one_r() replaces NULL values with
-				 * the NA value appropriate for the data type.
-				 */
-				pg_get_one_r(value, element_type, &result, idx);
+					isnull = elem_nulls[elem_idx];
+					itemvalue = elem_values[elem_idx++];
+
+					if (!isnull)
+					{
+						value = DatumGetCString(FunctionCall3(&out_func,
+															  itemvalue,
+															  (Datum) 0,
+															  Int32GetDatum(-1)));
+					}
+					else
+						value = NULL;
+
+					/*
+					 * Note that pg_get_one_r() replaces NULL values with
+					 * the NA value appropriate for the data type.
+					 */
+					pg_get_one_r(value, element_type, &result, idx);
+					if (value != NULL)
+						pfree(value);
+				}
 			}
 		}
-	}
-	UNPROTECT(1);
+		pfree(elem_values);
+		pfree(elem_nulls);
 
-	if (ndim > 1)
-	{
-		SEXP	matrix_dims;
+		if (ndim > 1)
+		{
+			SEXP	matrix_dims;
 
-		/* attach dimensions */
-		PROTECT(matrix_dims = allocVector(INTSXP, ndim));
-		for (i = 0; i < ndim; i++)
-			INTEGER_DATA(matrix_dims)[i] = dim[i];
+			/* attach dimensions */
+			PROTECT(matrix_dims = allocVector(INTSXP, ndim));
+			for (i = 0; i < ndim; i++)
+				INTEGER_DATA(matrix_dims)[i] = dim[i];
 
-		setAttrib(result, R_DimSymbol, matrix_dims);
-		UNPROTECT(1);
+			setAttrib(result, R_DimSymbol, matrix_dims);
+			UNPROTECT(1);
+		}
+
+		UNPROTECT(1);	/* result */
 	}
 
 	return result;
 }
 
 /*
- * given an array of pg tuples, convert to an R data.frame
+ * Given an array pg datums, convert to a multi-row R vector.
+ */
+SEXP
+pg_datum_array_get_r(Datum *elem_values, bool *elem_nulls, int numels, bool has_nulls,
+					 Oid element_type, FmgrInfo out_func, bool typbyval)
+{
+	/*
+	 * Loop through and convert each scalar value.
+	 * Use the converted values to build an R vector.
+	 */
+	SEXP		result;
+	int			i;
+	bool		fast_track_type;
+
+	switch (element_type)
+	{
+		case INT4OID:
+		case FLOAT8OID:
+			fast_track_type = true;
+			break;
+		default:
+			fast_track_type = false;
+	}
+
+	/*
+	 * Special case for pass-by-value data types, if the following conditions are met:
+	 * 		designated fast_track_type
+	 * 		no NULL elements
+	 * 		1 dimensional array only
+	 * 		at least one element
+	 */
+	if (fast_track_type &&
+		 typbyval &&
+		 !has_nulls &&
+		 (numels > 0))
+	{
+		SEXP	matrix_dims;
+
+		/* get new vector of the appropriate type and length */
+		PROTECT(result = get_r_vector(element_type, numels));
+
+		/* keep this in sync with switch above -- fast_track_type only */
+		switch (element_type)
+		{
+			case INT4OID:
+				Assert(sizeof(int) == 4);
+				memcpy(INTEGER_DATA(result), elem_values, numels * sizeof(int));
+				break;
+			case FLOAT8OID:
+				Assert(sizeof(double) == 8);
+				memcpy(NUMERIC_DATA(result), elem_values, numels * sizeof(double));
+				break;
+			default:
+				/* Everything else is error */
+				ereport(ERROR,
+						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+						 errmsg("direct array passthrough attempted for unsupported type")));
+		}
+
+		/* attach dimensions */
+		PROTECT(matrix_dims = allocVector(INTSXP, 1));
+		INTEGER_DATA(matrix_dims)[0] = numels;
+		setAttrib(result, R_DimSymbol, matrix_dims);
+		UNPROTECT(1);
+
+		UNPROTECT(1);	/* result */
+	}
+	else
+	{
+		SEXP	matrix_dims;
+
+		/* array is empty */
+		if (numels == 0)
+		{
+			PROTECT(result = get_r_vector(element_type, 0));
+			UNPROTECT(1);
+
+			return result;
+		}
+
+		/* get new vector of the appropriate type and length */
+		PROTECT(result = get_r_vector(element_type, numels));
+
+		/* Convert all values to their R form and build the vector */
+		for (i = 0; i < numels; i++)
+		{
+			char	   *value;
+			Datum		itemvalue;
+			bool		isnull;
+
+			isnull = elem_nulls[i];
+			itemvalue = elem_values[i];
+
+			if (!isnull)
+			{
+				value = DatumGetCString(FunctionCall3(&out_func,
+													  itemvalue,
+													  (Datum) 0,
+													  Int32GetDatum(-1)));
+			}
+			else
+				value = NULL;
+
+			/*
+			 * Note that pg_get_one_r() replaces NULL values with
+			 * the NA value appropriate for the data type.
+			 */
+			pg_get_one_r(value, element_type, &result, i);
+			if (value != NULL)
+				pfree(value);
+		}
+
+		/* attach dimensions */
+		PROTECT(matrix_dims = allocVector(INTSXP, 1));
+		INTEGER_DATA(matrix_dims)[0] = numels;
+		setAttrib(result, R_DimSymbol, matrix_dims);
+		UNPROTECT(1);
+
+		UNPROTECT(1);	/* result */
+	}
+
+	return result;
+}
+
+/*
+ * Given an array of pg tuples, convert to an R list
+ * the created object is not quite actually a data.frame
  */
 SEXP
 pg_tuple_get_r_frame(int ntuples, HeapTuple *tuples, TupleDesc tupdesc)
@@ -296,7 +495,7 @@ pg_tuple_get_r_frame(int ntuples, HeapTuple *tuples, TupleDesc tupdesc)
 		bool		typbyval;
 		char		typdelim;
 		Oid			typoutput,
-					elemtypelem;
+					typioparam;
 		FmgrInfo	outputproc;
 		char		typalign;
 
@@ -324,7 +523,7 @@ pg_tuple_get_r_frame(int ntuples, HeapTuple *tuples, TupleDesc tupdesc)
 		{
 			PROTECT(fldvec = NEW_LIST(nr));
 			get_type_io_data(typelem, IOFunc_output, &typlen, &typbyval,
-							 &typalign, &typdelim, &elemtypelem, &typoutput);
+							 &typalign, &typdelim, &typioparam, &typoutput);
 
 			fmgr_info(typoutput, &outputproc);
 		}
@@ -348,7 +547,10 @@ pg_tuple_get_r_frame(int ntuples, HeapTuple *tuples, TupleDesc tupdesc)
 				SEXP		fldvec_elem;
 
 				dvalue = SPI_getbinval(tuples[i], tupdesc, j + 1, &isnull);
-				PROTECT(fldvec_elem = pg_array_get_r(dvalue, outputproc, typlen, typbyval, typalign));
+				if (!isnull)
+					PROTECT(fldvec_elem = pg_array_get_r(dvalue, outputproc, typlen, typbyval, typalign));
+				else
+					PROTECT(fldvec_elem = R_NilValue);
 
 				SET_VECTOR_ELT(fldvec, i, fldvec_elem);
 				UNPROTECT(1);
@@ -372,7 +574,7 @@ pg_tuple_get_r_frame(int ntuples, HeapTuple *tuples, TupleDesc tupdesc)
 	}
 	setAttrib(result, R_RowNamesSymbol, row_names);
 
-	/* finally, tell R we are a "data.frame" */
+	/* finally, tell R we are a data.frame */
 	setAttrib(result, R_ClassSymbol, mkString("data.frame"));
 
 	UNPROTECT(3);
@@ -478,6 +680,17 @@ r_get_pg(SEXP rval, plr_function *function, FunctionCallInfo fcinfo)
 	bool	isnull = false;
 	Datum	result;
 
+	if (function->result_typid != BYTEAOID &&
+		(TYPEOF(rval) == CLOSXP ||
+		 TYPEOF(rval) == PROMSXP ||
+		 TYPEOF(rval) == LANGSXP ||
+		 TYPEOF(rval) == ENVSXP))
+		ereport(ERROR,
+				(errcode(ERRCODE_DATA_EXCEPTION),
+				 errmsg("incorrect function return type"),
+				 errdetail("R return value type cannot be mapped to PostgreSQL return type."),
+				 errhint("Try BYTEA as the PostgreSQL return type.")));
+
 	if (CALLED_AS_TRIGGER(fcinfo))
 		result = get_trigger_tuple(rval, function, fcinfo, &isnull);
 	else if (function->result_istuple || fcinfo->flinfo->fn_retset)
@@ -504,6 +717,30 @@ r_get_pg(SEXP rval, plr_function *function, FunctionCallInfo fcinfo)
 	return result;
 }
 
+/*
+ * Similar to r_get_pg, given an R value, convert to its pg representation
+ * Other than scalar, currently only prepared to be used with simple 1D vector
+ */
+Datum
+get_datum(SEXP rval, Oid typid, Oid typelem, FmgrInfo in_func, bool *isnull)
+{
+	Datum	result;
+
+	/* short circuit if return value is Null */
+	if (rval == R_NilValue || isNull(rval))	/* probably redundant */
+	{
+		*isnull = true;
+		return (Datum) 0;
+	}
+
+	if (typelem == InvalidOid)
+		result = get_scalar_datum(rval, typid, in_func, isnull);
+	else
+		result = get_simple_array_datum(rval, typelem, isnull);
+
+	return result;
+}
+
 static Datum
 get_trigger_tuple(SEXP rval, plr_function *function, FunctionCallInfo fcinfo, bool *isnull)
 {
@@ -669,7 +906,7 @@ get_trigger_tuple(SEXP rval, plr_function *function, FunctionCallInfo fcinfo, bo
 			if (values[j] != NULL)
 				pfree(values[j]);
 	}
-
+	UNPROTECT(1);
 	MemoryContextSwitchTo(oldcontext);
 
 	if (tuple)
@@ -896,9 +1133,11 @@ get_frame_array_datum(SEXP rval, plr_function *function, int col, bool *isnull)
 	ArrayType  *array;
 	int			nr = 0;
 	int			nc = length(rval);
-	int			ndims = 2;
-	int			dims[ndims];
-	int			lbs[ndims];
+#define FIXED_NUM_DIMS		2
+	int			ndims = FIXED_NUM_DIMS;
+	int			dims[FIXED_NUM_DIMS];
+	int			lbs[FIXED_NUM_DIMS];
+#undef FIXED_NUM_DIMS
 	int			idx;
 	SEXP		dfcol = NULL;
 	int			j;
@@ -939,7 +1178,13 @@ get_frame_array_datum(SEXP rval, plr_function *function, int col, bool *isnull)
 			/* internal error */
 			elog(ERROR, "plr: bad internal representation of data.frame");
 
-		if (ATTRIB(dfcol) == R_NilValue)
+		/*
+		 * Not sure about this test. Need to reliably detect
+		 * factors and do the alternative assignment ONLY for them.
+		 * For the moment this locution seems to work correctly.
+		 */
+		if (ATTRIB(dfcol) == R_NilValue ||
+			TYPEOF(CAR(ATTRIB(dfcol))) != STRSXP)
 			PROTECT(obj = AS_CHARACTER(dfcol));
 		else
 			PROTECT(obj = AS_CHARACTER(CAR(ATTRIB(dfcol))));
@@ -990,6 +1235,96 @@ get_frame_array_datum(SEXP rval, plr_function *function, int col, bool *isnull)
 	return dvalue;
 }
 
+/* return simple, one dimensional array */
+static Datum
+get_simple_array_datum(SEXP rval, Oid typelem, bool *isnull)
+{
+	Datum		dvalue;
+	SEXP		obj;
+	SEXP		rdims;
+	const char *value;
+	int16		typlen;
+	bool		typbyval;
+	char		typdelim;
+	Oid			typinput,
+				typioparam;
+	FmgrInfo	in_func;
+	char		typalign;
+	int			i;
+	Datum	   *dvalues = NULL;
+	ArrayType  *array;
+	int			nitems;
+	int		   *dims;
+	int		   *lbs;
+	bool	   *nulls;
+	bool		have_nulls = FALSE;
+	int			ndims = 1;
+
+	dims = palloc(ndims * sizeof(int));
+	lbs = palloc(ndims * sizeof(int));
+
+	/*
+	 * get the element type's in_func
+	 */
+	get_type_io_data(typelem, IOFunc_output, &typlen, &typbyval,
+					 &typalign, &typdelim, &typioparam, &typinput);
+
+	perm_fmgr_info(typinput, &in_func);
+
+	PROTECT(rdims = getAttrib(rval, R_DimSymbol));
+	if (length(rdims) > 1)
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("greater than 1-dimensional arrays are " \
+						"not supported in this context")));
+
+	dims[0] = INTEGER(rdims)[0];
+	lbs[0] = 1;
+	UNPROTECT(1);
+
+	nitems = dims[0];
+	if (nitems == 0)
+	{
+		*isnull = true;
+		return (Datum) 0;
+	}
+
+	dvalues = (Datum *) palloc(nitems * sizeof(Datum));
+	nulls = (bool *) palloc(nitems * sizeof(bool));
+	PROTECT(obj =  AS_CHARACTER(rval));
+
+	for (i = 0; i < nitems; i++)
+	{
+		value = CHAR(STRING_ELT(obj, i));
+
+		if (STRING_ELT(obj, i) == NA_STRING || value == NULL)
+		{
+			nulls[i] = TRUE;
+			have_nulls = TRUE;
+		}
+		else
+		{
+			nulls[i] = FALSE;
+			dvalues[i] = FunctionCall3(&in_func,
+										CStringGetDatum(value),
+										(Datum) 0,
+										Int32GetDatum(-1));
+		}
+	}
+	UNPROTECT(1);
+
+	if (!have_nulls)
+		array = construct_md_array(dvalues, NULL, ndims, dims, lbs,
+									typelem, typlen, typbyval, typalign);
+	else
+		array = construct_md_array(dvalues, nulls, ndims, dims, lbs,
+									typelem, typlen, typbyval, typalign);
+
+	dvalue = PointerGetDatum(array);
+
+	return dvalue;
+}
+
 static Datum
 get_md_array_datum(SEXP rval, int ndims, plr_function *function, int col, bool *isnull)
 {
@@ -1009,13 +1344,24 @@ get_md_array_datum(SEXP rval, int ndims, plr_function *function, int col, bool *
 	int			nr = 1;
 	int			nc = 1;
 	int			nz = 1;
-	int			dims[ndims];
-	int			lbs[ndims];
+	int		   *dims;
+	int		   *lbs;
 	int			idx;
 	int			cntr = 0;
 	bool	   *nulls;
 	bool		have_nulls = FALSE;
 
+	if (ndims > 0)
+	{
+		dims = palloc(ndims * sizeof(int));
+		lbs = palloc(ndims * sizeof(int));
+	}
+	else
+	{
+		dims = NULL;
+		lbs = NULL;
+	}
+	
 	if (function->result_istuple)
 	{
 		result_elem = function->result_fld_elem_typid[col];
@@ -1122,12 +1468,16 @@ get_generic_array_datum(SEXP rval, plr_function *function, int col, bool *isnull
 	int			i;
 	Datum	   *dvalues = NULL;
 	ArrayType  *array;
-	int			ndims = 1;
-	int			dims[ndims];
-	int			lbs[ndims];
+#define FIXED_NUM_DIMS		1
+	int			ndims = FIXED_NUM_DIMS;
+	int			dims[FIXED_NUM_DIMS];
+	int			lbs[FIXED_NUM_DIMS];
+#undef FIXED_NUM_DIMS
 	bool	   *nulls;
 	bool		have_nulls = FALSE;
-	
+	bool		fast_track_type;
+	bool		has_na = false;
+
 	if (function->result_istuple)
 	{
 		result_elem = function->result_fld_elem_typid[col];
@@ -1144,43 +1494,132 @@ get_generic_array_datum(SEXP rval, plr_function *function, int col, bool *isnull
 		typbyval = function->result_elem_typbyval;
 		typalign = function->result_elem_typalign;
 	}
-	
-	dvalues = (Datum *) palloc(objlen * sizeof(Datum));
-	nulls = (bool *) palloc(objlen * sizeof(bool));
-	PROTECT(obj =  AS_CHARACTER(rval));
 
-	/* Loop is needed here as result value might be of length > 1 */
-	for(i = 0; i < objlen; i++)
+	/*
+	 * Special case for pass-by-value data types, if the following conditions are met:
+	 * 		designated fast_track_type
+	 * 		no NULL/NA elements
+	 */
+	if (TYPEOF(rval) == INTSXP ||
+		TYPEOF(rval) == REALSXP)
 	{
-		value = CHAR(STRING_ELT(obj, i));
+		switch (TYPEOF(rval)) {
+		case INTSXP:
+			if (result_elem == INT4OID)
+				fast_track_type = true;
+			else
+				fast_track_type = false;
 
-		if (STRING_ELT(obj, i) == NA_STRING || value == NULL)
-		{
-			nulls[i] = TRUE;
-			have_nulls = TRUE;
-		}
-		else
-		{
-			nulls[i] = FALSE;
-			dvalues[i] = FunctionCall3(&in_func,
-									CStringGetDatum(value),
-									(Datum) 0,
-									Int32GetDatum(-1));
+			for (i = 0; i < objlen; i++)
+			{
+				if (INTEGER(rval)[i] == NA_INTEGER)
+				{
+					has_na = true;
+					break;
+				}
+			}
+			break;
+		case REALSXP:
+			if (result_elem == FLOAT8OID)
+				fast_track_type = true;
+			else
+				fast_track_type = false;
+
+			for (i = 0; i < objlen; i++)
+			{
+				if (ISNAN(REAL(rval)[i]))
+				{
+					has_na = true;
+					break;
+				}
+			}
+			break;
+		default:
+			fast_track_type = false;
+			has_na = true;	/* does not really matter in this case */
 		}
 	}
-	UNPROTECT(1);
-
-	dims[0] = objlen;
-	lbs[0] = 1;
-
-	if (!have_nulls)
-		array = construct_md_array(dvalues, NULL, ndims, dims, lbs,
-									result_elem, typlen, typbyval, typalign);
 	else
-		array = construct_md_array(dvalues, nulls, ndims, dims, lbs,
-									result_elem, typlen, typbyval, typalign);
+	{
+		fast_track_type = false;
+		has_na = true;	/* does not really matter in this case */
+	}
 
-	dvalue = PointerGetDatum(array);
+	if (fast_track_type &&
+		 typbyval &&
+		 !has_na)
+	{
+		int32		nbytes = 0;
+		int32		dataoffset;
+
+		if (TYPEOF(rval) == INTSXP)
+		{
+			nbytes = objlen * sizeof(INTEGER_DATA(rval));
+			dvalues = (Datum *) INTEGER_DATA(rval);
+		}
+		else if (TYPEOF(rval) == REALSXP)
+		{
+			nbytes = objlen * sizeof(NUMERIC_DATA(rval));
+			dvalues = (Datum *) NUMERIC_DATA(rval);
+		}
+		else
+			elog(ERROR, "attempted to passthrough invalid R datatype to Postgresql");
+
+		dims[0] = objlen;
+		lbs[0] = 1;
+		dataoffset = 0;			/* marker for no null bitmap */
+
+		array = (ArrayType *) palloc0(nbytes + ARR_OVERHEAD_NONULLS(ndims));
+		SET_VARSIZE(array, nbytes + ARR_OVERHEAD_NONULLS(ndims));
+		array->ndim = ndims;
+		array->dataoffset = dataoffset;
+		array->elemtype = result_elem;
+		memcpy(ARR_DIMS(array), dims, ndims * sizeof(int));
+		memcpy(ARR_LBOUND(array), lbs, ndims * sizeof(int));
+		memcpy(ARR_DATA_PTR(array), dvalues, nbytes);
+
+		dvalue = PointerGetDatum(array);
+	}
+	else
+	{
+		/* original code */
+		dvalues = (Datum *) palloc(objlen * sizeof(Datum));
+		nulls = (bool *) palloc(objlen * sizeof(bool));
+		PROTECT(obj =  AS_CHARACTER(rval));
+
+		/* Loop is needed here as result value might be of length > 1 */
+		for(i = 0; i < objlen; i++)
+		{
+			value = CHAR(STRING_ELT(obj, i));
+
+			if (STRING_ELT(obj, i) == NA_STRING || value == NULL)
+			{
+				nulls[i] = TRUE;
+				have_nulls = TRUE;
+			}
+			else
+			{
+				nulls[i] = FALSE;
+				dvalues[i] = FunctionCall3(&in_func,
+										CStringGetDatum(value),
+										(Datum) 0,
+										Int32GetDatum(-1));
+			}
+		}
+		UNPROTECT(1);
+
+		dims[0] = objlen;
+		lbs[0] = 1;
+
+		if (!have_nulls)
+			array = construct_md_array(dvalues, NULL, ndims, dims, lbs,
+										result_elem, typlen, typbyval, typalign);
+		else
+			array = construct_md_array(dvalues, nulls, ndims, dims, lbs,
+										result_elem, typlen, typbyval, typalign);
+
+		dvalue = PointerGetDatum(array);
+	}
 
 	return dvalue;
 }
@@ -1243,7 +1682,9 @@ get_frame_tuplestore(SEXP rval,
 	for (j = 0; j < nc; j++)
 	{
 		PROTECT(dfcol = VECTOR_ELT(rval, j));
-		if(!isFactor(dfcol))
+		if((!isFactor(dfcol)) &&
+		   ((attrs[j]->attndims == 0) ||
+			(TYPEOF(dfcol) != VECSXP)))
 		{
 			SEXP	obj;
 
@@ -1251,6 +1692,25 @@ get_frame_tuplestore(SEXP rval,
 			SET_VECTOR_ELT(result, j, obj);
 			UNPROTECT(1);
 		}
+		else if(attrs[j]->attndims != 0)	/* array data type */
+		{
+			SEXP	obj;
+
+			PROTECT(obj = NEW_LIST(nr));
+			for(i = 0; i < nr; i++)
+			{
+				SEXP	dfcolcell;
+				SEXP	objcell;
+
+				PROTECT(dfcolcell = VECTOR_ELT(dfcol, i));
+				PROTECT(objcell = AS_CHARACTER(dfcolcell));
+				SET_VECTOR_ELT(obj, i, objcell);
+				UNPROTECT(2);
+			}
+
+			SET_VECTOR_ELT(result, j, obj);
+			UNPROTECT(1);
+		}
 		else
 		{
 			SEXP 	t;
@@ -1308,18 +1768,26 @@ get_frame_tuplestore(SEXP rval,
 			}
 			else
 			{
-				if (STRING_ELT(dfcol, i) != NA_STRING)
+				if ((attrs[j]->attndims != 0) || (STRING_ELT(dfcol, i) != NA_STRING))
 				{
 					if (attrs[j]->attndims == 0)
+					{
 						values[j] = pstrdup(CHAR(STRING_ELT(dfcol, i)));
-					else
+					}
+					else	/* array data type */
 					{
 						bool	isnull = false;
 						Datum	arr_datum;
-						
-						arr_datum = get_array_datum(dfcol, function, j, &isnull);
+
+						if (TYPEOF(dfcol) != VECSXP)
+							arr_datum = get_array_datum(dfcol, function, j, &isnull);
+						else
+							arr_datum = get_array_datum(VECTOR_ELT(dfcol,i), function, j, &isnull);
+
 						if (isnull)
+						{
 							values[j] = NULL;
+						}
 						else
 						{
 							FunctionCallInfoData	fake_fcinfo;
