@@ -2,7 +2,7 @@
  * PL/R - PostgreSQL support for R as a
  *	      procedural language (PL)
  *
- * Copyright (c) 2003-2009 by Joseph E. Conway
+ * Copyright (c) 2003-2010 by Joseph E. Conway
  * ALL RIGHTS RESERVED
  * 
  * Joe Conway <mail@joeconway.com>
@@ -31,7 +31,6 @@
  * pg_rsupport.c - Postgres support for use within plr functions
  */
 #include "plr.h"
-#include "utils/memutils.h"
 
 extern MemoryContext plr_SPI_context;
 extern char *last_R_error_msg;
@@ -147,7 +146,7 @@ plr_SPI_exec(SEXP rsql)
 		error("%s", "cannot exec empty query");
 
 	/* switch to SPI memory context */
-	oldcontext = MemoryContextSwitchTo(plr_SPI_context);
+	SWITCHTO_PLR_SPI_CONTEXT(oldcontext);
 
 	/*
 	 * trap elog/ereport so we can let R finish up gracefully
@@ -331,7 +330,7 @@ plr_SPI_prepare(SEXP rsql, SEXP rargtypes)
 
 			get_type_io_data(typeids[i], IOFunc_input, &typlen, &typbyval,
 							 &typalign, &typdelim, &typelem, &typinput);
-			typelems[i] = typelem;
+			typelems[i] = get_element_type(typeids[i]);
 
 			MemoryContextSwitchTo(oldcontext);
 
@@ -346,7 +345,7 @@ plr_SPI_prepare(SEXP rsql, SEXP rargtypes)
 	UNPROTECT(1);
 
 	/* switch to SPI memory context */
-	oldcontext = MemoryContextSwitchTo(plr_SPI_context);
+	SWITCHTO_PLR_SPI_CONTEXT(oldcontext);
 
 	/*
 	 * trap elog/ereport so we can let R finish up gracefully
@@ -456,6 +455,7 @@ plr_SPI_execp(SEXP rsaved_plan, SEXP rargvalues)
 	void			   *saved_plan = plan_desc->saved_plan;
 	int					nargs = plan_desc->nargs;
 	Oid				   *typeids = plan_desc->typeids;
+	Oid				   *typelems = plan_desc->typelems;
 	FmgrInfo		   *typinfuncs = plan_desc->typinfuncs;
 	int					i;
 	Datum			   *argvalues = NULL;
@@ -492,7 +492,7 @@ plr_SPI_execp(SEXP rsaved_plan, SEXP rargvalues)
 	{
 		PROTECT(obj = VECTOR_ELT(rargvalues, i));
 
-		argvalues[i] = get_scalar_datum(obj, typeids[i], typinfuncs[i], &isnull);
+		argvalues[i] = get_datum(obj, typeids[i], typelems[i], typinfuncs[i], &isnull);
 		if (!isnull)
 			nulls[i] = ' ';
 		else
@@ -502,7 +502,7 @@ plr_SPI_execp(SEXP rsaved_plan, SEXP rargvalues)
 	}
 
 	/* switch to SPI memory context */
-	oldcontext = MemoryContextSwitchTo(plr_SPI_context);
+	SWITCHTO_PLR_SPI_CONTEXT(oldcontext);
 
 	/*
 	 * trap elog/ereport so we can let R finish up gracefully
@@ -660,7 +660,7 @@ plr_SPI_cursor_open(SEXP cursor_name_arg,SEXP rsaved_plan, SEXP rargvalues)
 	strncpy(cursor_name, CHAR(STRING_ELT(cursor_name_arg,0)), 64);
 
 	/* switch to SPI memory context */
-	oldcontext = MemoryContextSwitchTo(plr_SPI_context);
+	SWITCHTO_PLR_SPI_CONTEXT(oldcontext);
 
 	/*
 	 * trap elog/ereport so we can let R finish up gracefully
@@ -714,7 +714,7 @@ plr_SPI_cursor_fetch(SEXP cursor_in,SEXP forward_in, SEXP rows_in)
 	rows  = INTEGER_DATA(rows_in)[0];
 
 	/* switch to SPI memory context */
-	oldcontext = MemoryContextSwitchTo(plr_SPI_context);
+	SWITCHTO_PLR_SPI_CONTEXT(oldcontext);
 	PG_TRY();
 	{
 		/* Open the cursor */
@@ -751,7 +751,7 @@ plr_SPI_cursor_close(SEXP cursor_in)
 	portal = R_ExternalPtrAddr(cursor_in);
 
 	/* switch to SPI memory context */
-	oldcontext = MemoryContextSwitchTo(plr_SPI_context);
+	SWITCHTO_PLR_SPI_CONTEXT(oldcontext);
 	PG_TRY();
 	{
 		/* Open the cursor */
@@ -788,7 +788,7 @@ plr_SPI_cursor_move(SEXP cursor_in,SEXP forward_in, SEXP rows_in)
 	rows  = INTEGER(rows_in)[0];
 
 	/* switch to SPI memory context */
-	oldcontext = MemoryContextSwitchTo(plr_SPI_context);
+	SWITCHTO_PLR_SPI_CONTEXT(oldcontext);
 	PG_TRY();
 	{
 		/* Open the cursor */
