@@ -2,7 +2,7 @@
  * PL/R - PostgreSQL support for R as a
  *	      procedural language (PL)
  *
- * Copyright (c) 2003-2009 by Joseph E. Conway
+ * Copyright (c) 2003-2010 by Joseph E. Conway
  * ALL RIGHTS RESERVED
  * 
  * Joe Conway <mail@joeconway.com>
@@ -31,18 +31,15 @@
  * plr.c - Language handler and support functions
  */
 #include "plr.h"
-#include "catalog/pg_namespace.h"
-#include "storage/ipc.h"
-#include "utils/memutils.h"
-#include "utils/typcache.h"
+#include "libpq/pqsignal.h"
 
 PG_MODULE_MAGIC;
 
 /*
  * Global data
  */
-MemoryContext plr_SPI_context;
 MemoryContext plr_caller_context;
+MemoryContext plr_SPI_context = NULL;
 HTAB *plr_HashTable = (HTAB *) NULL;
 char *last_R_error_msg = NULL;
 
@@ -174,12 +171,15 @@ static plr_function *do_compile(FunctionCallInfo fcinfo,
 								HeapTuple procTup,
 								plr_func_hashkey *hashkey);
 static SEXP plr_parse_func_body(const char *body);
-static SEXP plr_convertargs(plr_function *function, Datum *arg, bool *argnull);
+static SEXP plr_convertargs(plr_function *function, Datum *arg, bool *argnull, FunctionCallInfo fcinfo);
 static void plr_error_callback(void *arg);
 static Oid getNamespaceOidFromFunctionOid(Oid fnOid);
 static bool haveModulesTable(Oid nspOid);
 static char *getModulesSql(Oid nspOid);
 static char **fetchArgNames(HeapTuple procTup, int nargs);
+#ifdef HAVE_WINDOW_FUNCTIONS
+static void WinGetFrameData(WindowObject winobj, int argno, Datum *dvalues, bool *isnull, int *numels, bool *has_nulls);
+#endif
 
 /*
  * plr_call_handler -	This is the only visible function
@@ -194,20 +194,14 @@ Datum
 plr_call_handler(PG_FUNCTION_ARGS)
 {
 	Datum			retval;
-	MemoryContext	origcontext = CurrentMemoryContext;
-	/* save caller's SPI context */
-	MemoryContext	plr_SPI_context_save = plr_SPI_context;
+
 	/* save caller's context */
-	plr_caller_context = origcontext;
+	plr_caller_context = CurrentMemoryContext;
 
-	/* Connect to SPI manager */
 	if (SPI_connect() != SPI_OK_CONNECT)
-		ereport(ERROR,
-				(errcode(ERRCODE_CONNECTION_FAILURE),
-				 errmsg("cannot connect to SPI manager")));
-
-	/* switch from SPI back to original call memory context */
-	plr_SPI_context = MemoryContextSwitchTo(origcontext);
+		elog(ERROR, "SPI_connect failed");
+	plr_SPI_context = CurrentMemoryContext;
+	MemoryContextSwitchTo(plr_caller_context);
 
 	/* initialize R if needed */
 	plr_init_all(fcinfo->flinfo->fn_oid);
@@ -217,17 +211,6 @@ plr_call_handler(PG_FUNCTION_ARGS)
 	else
 		retval = plr_func_handler(fcinfo);
 
-	/* switch back to SPI memory context */
-	MemoryContextSwitchTo(plr_SPI_context);
-
-	if (SPI_finish() != SPI_OK_FINISH)
-		ereport(ERROR,
-				(errcode(ERRCODE_CONNECTION_EXCEPTION),
-				 errmsg("SPI_finish() failed")));
-
-	/* restore caller's SPI context */
-	plr_SPI_context = plr_SPI_context_save;
-
 	return retval;
 }
 
@@ -235,8 +218,7 @@ void
 load_r_cmd(const char *cmd)
 {
 	SEXP		cmdSexp,
-				cmdexpr,
-				ans = R_NilValue;
+				cmdexpr;
 	int			i,
 				status;
 
@@ -268,7 +250,7 @@ load_r_cmd(const char *cmd)
 	/* Loop is needed here as EXPSEXP may be of length > 1 */
 	for(i = 0; i < length(cmdexpr); i++)
 	{
-		ans = R_tryEval(VECTOR_ELT(cmdexpr, i), R_GlobalEnv, &status);
+		R_tryEval(VECTOR_ELT(cmdexpr, i), R_GlobalEnv, &status);
 		if(status != 0)
 		{
 			if (last_R_error_msg)
@@ -313,6 +295,8 @@ PLR_CLEANUP
 		 */
 		buf = (char *) palloc(9 + 1 + strlen(tmpdir));
 		sprintf(buf, "rm -rf \"%s\"", tmpdir);
+
+		/* ignoring return value */
 		system(buf);
 	}
 }
@@ -342,7 +326,7 @@ plr_init(void)
 {
 	char	   *r_home;
 	int			rargc;
-	char	   *rargv[] = {"PL/R", "--silent", "--no-save", "--no-restore"};
+	char	   *rargv[] = {"PL/R", "--slave", "--silent", "--no-save", "--no-restore"};
 
 	/* refuse to init more than once */
 	if (plr_pm_init_done)
@@ -408,6 +392,11 @@ plr_init(void)
 	R_Interactive = false;
 #endif
 
+	/*
+	 * R seems to try to steal SIGINT in recent releases, so steal it back
+	 */
+	pqsignal(SIGINT, StatementCancelHandler);		/* cancel current query */
+
 	plr_pm_init_done = true;
 }
 
@@ -485,7 +474,7 @@ plr_load_builtins(Oid funcid)
  * DO NOT make this static --- it has to be callable by reload_plr_modules()
  */
 void
-plr_load_modules(MemoryContext	plr_SPI_context)
+plr_load_modules(void)
 {
 	int				spi_rc;
 	char		   *cmd;
@@ -495,13 +484,17 @@ plr_load_modules(MemoryContext	plr_SPI_context)
 	char		   *modulesSql;
 
 	/* switch to SPI memory context */
-	oldcontext = MemoryContextSwitchTo(plr_SPI_context);
+	SWITCHTO_PLR_SPI_CONTEXT(oldcontext);
 
 	/*
 	 * Check if table plr_modules exists
 	 */
 	if (!haveModulesTable(plr_nspOid))
+	{
+		/* clean up if SPI was used, and regardless restore caller's context */
+		CLEANUP_PLR_SPI_CONTEXT(oldcontext);
 		return;
+	}
 
 	/* plr_modules table exists -- get SQL code extract table's contents */
 	modulesSql = getModulesSql(plr_nspOid);
@@ -520,6 +513,8 @@ plr_load_modules(MemoryContext	plr_SPI_context)
 	if (SPI_processed == 0)
 	{
 		SPI_freetuptable(SPI_tuptable);
+		/* clean up if SPI was used, and regardless restore caller's context */
+		CLEANUP_PLR_SPI_CONTEXT(oldcontext);
 		return;
 	}
 
@@ -542,8 +537,8 @@ plr_load_modules(MemoryContext	plr_SPI_context)
 	}
 	SPI_freetuptable(SPI_tuptable);
 
-	/* back to caller's memory context */
-	MemoryContextSwitchTo(oldcontext);
+	/* clean up if SPI was used, and regardless restore caller's context */
+	CLEANUP_PLR_SPI_CONTEXT(oldcontext);
 }
 
 static void
@@ -571,7 +566,7 @@ plr_init_all(Oid funcid)
 		plr_nspOid = getNamespaceOidFromFunctionOid(funcid);
 
 		/* try to load procedures from plr_modules */
-		plr_load_modules(plr_SPI_context);
+		plr_load_modules();
 
 		plr_be_init_done = true;
 	}
@@ -592,15 +587,22 @@ plr_trigger_handler(PG_FUNCTION_ARGS)
 	bool			argnull[FUNC_MAX_ARGS];
 	TriggerData	   *trigdata = (TriggerData *) fcinfo->context;
 	TupleDesc		tupdesc = trigdata->tg_relation->rd_att;
-	Datum			dvalues[trigdata->tg_trigger->tgnargs];
+	Datum		   *dvalues;
 	ArrayType	   *array;
-	int				ndims = 1;
-	int				dims[ndims];
-	int				lbs[ndims];
+#define FIXED_NUM_DIMS		1
+	int				ndims = FIXED_NUM_DIMS;
+	int				dims[FIXED_NUM_DIMS];
+	int				lbs[FIXED_NUM_DIMS];
+#undef FIXED_NUM_DIMS
 	TRIGGERTUPLEVARS;
 	ERRORCONTEXTCALLBACK;
 	int				i;
 
+	if (trigdata->tg_trigger->tgnargs > 0)
+		dvalues = palloc(trigdata->tg_trigger->tgnargs * sizeof(Datum));
+	else
+		dvalues = NULL;
+	
 	/* Find or compile the function */
 	function = compile_plr_function(fcinfo);
 
@@ -708,7 +710,7 @@ plr_trigger_handler(PG_FUNCTION_ARGS)
 	PROTECT(fun = function->fun);
 
 	/* Convert all call arguments */
-	PROTECT(rargs = plr_convertargs(function, arg, argnull));
+	PROTECT(rargs = plr_convertargs(function, arg, argnull, fcinfo));
 
 	/* Call the R function */
 	PROTECT(rvalue = call_r_func(fun, rargs));
@@ -717,6 +719,8 @@ plr_trigger_handler(PG_FUNCTION_ARGS)
 	 * Convert the return value from an R object to a Datum.
 	 * We expect r_get_pg to do the right thing with missing or empty results.
 	 */
+	if (SPI_finish() != SPI_OK_FINISH)
+		elog(ERROR, "SPI_finish failed");
 	retval = r_get_pg(rvalue, function, fcinfo);
 
 	POP_PLERRCONTEXT;
@@ -744,7 +748,7 @@ plr_func_handler(PG_FUNCTION_ARGS)
 	PROTECT(fun = function->fun);
 
 	/* Convert all call arguments */
-	PROTECT(rargs = plr_convertargs(function, fcinfo->arg, fcinfo->argnull));
+	PROTECT(rargs = plr_convertargs(function, fcinfo->arg, fcinfo->argnull, fcinfo));
 
 	/* Call the R function */
 	PROTECT(rvalue = call_r_func(fun, rargs));
@@ -753,6 +757,8 @@ plr_func_handler(PG_FUNCTION_ARGS)
 	 * Convert the return value from an R object to a Datum.
 	 * We expect r_get_pg to do the right thing with missing or empty results.
 	 */
+	if (SPI_finish() != SPI_OK_FINISH)
+		elog(ERROR, "SPI_finish failed");
 	retval = r_get_pg(rvalue, function, fcinfo);
 
 	POP_PLERRCONTEXT;
@@ -940,6 +946,11 @@ do_compile(FunctionCallInfo fcinfo,
 	function->fn_xmin = HeapTupleHeaderGetXmin(procTup->t_data);
 	function->fn_tid = procTup->t_self;
 
+#ifdef HAVE_WINDOW_FUNCTIONS
+	/* Flag for window functions */
+	function->iswindow = procStruct->proiswindow;
+#endif
+
 	/* Lookup the pg_language tuple by Oid*/
 	langTup = SearchSysCache(LANGOID,
 							 ObjectIdGetDatum(procStruct->prolang),
@@ -1083,8 +1094,8 @@ do_compile(FunctionCallInfo fcinfo,
 				function->result_elem = InvalidOid;
 			
 			/*
-			* if we have an array type, get the element type's in_func
-			*/
+			 * if we have an array type, get the element type's in_func
+			 */
 			if (function->result_elem != InvalidOid)
 			{
 				int16		typlen;
@@ -1140,24 +1151,28 @@ do_compile(FunctionCallInfo fcinfo,
 									 0, 0, 0);
 			if (!HeapTupleIsValid(typeTup))
 			{
+				Oid		arg_typid = function->arg_typid[i];
+
 				xpfree(function->proname);
 				xpfree(function);
 				/* internal error */
-				elog(ERROR, "cache lookup failed for argument type %u",
-					 function->arg_typid[i]);
+				elog(ERROR, "cache lookup failed for argument type %u", arg_typid);
 			}
 			typeStruct = (Form_pg_type) GETSTRUCT(typeTup);
 
-			/* Disallow pseudotype argument */
-			/* (note we already replaced ANYARRAY/ANYELEMENT) */
+			/* Disallow pseudotype argument
+			 * note we already replaced ANYARRAY/ANYELEMENT
+			 */
 			if (typeStruct->typtype == 'p')
 			{
+				Oid		arg_typid = function->arg_typid[i];
+
 				xpfree(function->proname);
 				xpfree(function);
 				ereport(ERROR,
 						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 						 errmsg("plr functions cannot take type %s",
-								format_type_be(function->arg_typid[i]))));
+								format_type_be(arg_typid))));
 			}
 
 			if (typeStruct->typrelid != InvalidOid)
@@ -1167,6 +1182,9 @@ do_compile(FunctionCallInfo fcinfo,
 
 			perm_fmgr_info(typeStruct->typoutput, &(function->arg_out_func[i]));
 
+			/* save argument typbyval in case we need for optimization in conversions */
+			function->arg_typbyval[i] = typeStruct->typbyval;
+
 			/*
 			 * Is argument type an array? get_element_type will return InvalidOid
 			 * instead of actual element type if the type is not a varlena array.
@@ -1205,6 +1223,18 @@ do_compile(FunctionCallInfo fcinfo,
 			}
 		}
 		FREE_ARG_NAMES;
+
+#ifdef HAVE_WINDOW_FUNCTIONS
+		if (function->iswindow)
+		{
+			for (i = 0; i < function->nargs; i++)
+			{
+				appendStringInfo(proc_internal_args, ",");
+				SET_FRAME_ARG_NAME;
+			}
+			SET_FRAME_XARG_NAMES;
+		}
+#endif
 	}
 	else
 	{
@@ -1443,17 +1473,31 @@ call_r_func(SEXP fun, SEXP rargs)
 }
 
 static SEXP
-plr_convertargs(plr_function *function, Datum *arg, bool *argnull)
+plr_convertargs(plr_function *function, Datum *arg, bool *argnull, FunctionCallInfo fcinfo)
 {
 	int		i;
+	int		m = 1;
+	int		c = 0;
 	SEXP	rargs,
 			el;
 
+#ifdef HAVE_WINDOW_FUNCTIONS
+	if (function->iswindow)
+	{
+		/*
+		 * For WINDOW functions, create an array of R objects with
+		 * the number of elements equal to twice the number of arguments.
+		 */
+		m = 2;
+		c = 2;
+	}
+#endif
+
 	/*
 	 * Create an array of R objects with the number of elements
-	 * equal to the number of arguments.
+	 * as a function of the number of arguments.
 	 */
-	PROTECT(rargs = allocVector(VECSXP, function->nargs));
+	PROTECT(rargs = allocVector(VECSXP, c + (m * function->nargs)));
 
 	/*
 	 * iterate over the arguments, convert each of them and put them in
@@ -1461,43 +1505,139 @@ plr_convertargs(plr_function *function, Datum *arg, bool *argnull)
 	 */
 	for (i = 0; i < function->nargs; i++)
 	{
-		if (argnull[i])
+#ifdef HAVE_WINDOW_FUNCTIONS
+		if (!function->iswindow)
 		{
-			/* fast track for null arguments */
-			PROTECT(el = R_NilValue);
-		}
-		else if (function->arg_is_rel[i])
-		{
-			/* for tuple args, convert to a one row data.frame */
-			CONVERT_TUPLE_TO_DATAFRAME;
-		}
-		else if (function->arg_elem[i] == InvalidOid)
-		{
-			/* for scalar args, convert to a one row vector */
-			Datum		dvalue = arg[i];
-			Oid			arg_typid = function->arg_typid[i];
-			FmgrInfo	arg_out_func = function->arg_out_func[i];
+#endif
+			if (argnull[i])
+			{
+				/* fast track for null arguments */
+				PROTECT(el = R_NilValue);
+			}
+			else if (function->arg_is_rel[i])
+			{
+				/* for tuple args, convert to a one row data.frame */
+				CONVERT_TUPLE_TO_DATAFRAME;
+			}
+			else if (function->arg_elem[i] == InvalidOid)
+			{
+				/* for scalar args, convert to a one row vector */
+				Datum		dvalue = arg[i];
+				Oid			arg_typid = function->arg_typid[i];
+				FmgrInfo	arg_out_func = function->arg_out_func[i];
 
-			PROTECT(el = pg_scalar_get_r(dvalue, arg_typid, arg_out_func));
+				PROTECT(el = pg_scalar_get_r(dvalue, arg_typid, arg_out_func));
+			}
+			else
+			{
+				/* better be a pg array arg, convert to a multi-row vector */
+				Datum		dvalue = (Datum) PG_DETOAST_DATUM(arg[i]);
+				FmgrInfo	out_func = function->arg_elem_out_func[i];
+				int			typlen = function->arg_elem_typlen[i];
+				bool		typbyval = function->arg_elem_typbyval[i];
+				char		typalign = function->arg_elem_typalign[i];
+
+				PROTECT(el = pg_array_get_r(dvalue, out_func, typlen, typbyval, typalign));
+			}
+			SET_VECTOR_ELT(rargs, i, el);
+			UNPROTECT(1);
+#ifdef HAVE_WINDOW_FUNCTIONS
 		}
 		else
 		{
-			/* better be a pg array arg, convert to a multi-row vector */
-			Datum		dvalue = (Datum) PG_DETOAST_DATUM(arg[i]);
-			FmgrInfo	out_func = function->arg_elem_out_func[i];
-			int			typlen = function->arg_elem_typlen[i];
-			bool		typbyval = function->arg_elem_typbyval[i];
-			char		typalign = function->arg_elem_typalign[i];
+			Datum			dvalue;
+			bool			isnull;
+			WindowObject	winobj = PG_WINDOW_OBJECT();
 
-			PROTECT(el = pg_array_get_r(dvalue, out_func, typlen, typbyval, typalign));
+			/* get datum for the current row of the window frame */
+			dvalue = WinGetFuncArgInFrame(winobj, i, 0, WINDOW_SEEK_CURRENT, false, &isnull, NULL);
+
+			if (isnull)
+			{
+				/* fast track for null arguments */
+				PROTECT(el = R_NilValue);
+			}
+			else if (function->arg_is_rel[i])
+			{
+				/* keep compiler quiet */
+				el = R_NilValue;
+
+				elog(ERROR, "Tuple arguments not supported in PL/R Window Functions");
+			}
+			else if (function->arg_elem[i] == InvalidOid)
+			{
+				/* for scalar args, convert to a one row vector */
+				Oid			arg_typid = function->arg_typid[i];
+				FmgrInfo	arg_out_func = function->arg_out_func[i];
+
+				PROTECT(el = pg_scalar_get_r(dvalue, arg_typid, arg_out_func));
+			}
+			else
+			{
+				/* better be a pg array arg, convert to a multi-row vector */
+				FmgrInfo	out_func = function->arg_elem_out_func[i];
+				int			typlen = function->arg_elem_typlen[i];
+				bool		typbyval = function->arg_elem_typbyval[i];
+				char		typalign = function->arg_elem_typalign[i];
+
+				dvalue = (Datum) PG_DETOAST_DATUM(dvalue);
+				PROTECT(el = pg_array_get_r(dvalue, out_func, typlen, typbyval, typalign));
+			}
+			SET_VECTOR_ELT(rargs, i, el);
+			UNPROTECT(1);
 		}
-
-		SET_VECTOR_ELT(rargs, i, el);
-		UNPROTECT(1);
+#endif
 	}
 
-	UNPROTECT(1);
+#ifdef HAVE_WINDOW_FUNCTIONS
+	/* now get an array of datums for the entire window frame for each argument */
+	if (function->iswindow)
+	{
+		WindowObject	winobj = PG_WINDOW_OBJECT();
+		int64			totalrows = WinGetPartitionRowCount(winobj);
+		int				numels = 0;
 
+		for (i = 0; i < function->nargs; i++)
+		{
+			Datum		   *dvalues = palloc0(totalrows * sizeof(Datum));
+			bool		   *isnulls = palloc0(totalrows * sizeof(bool));
+			Oid				datum_typid;
+			FmgrInfo		datum_out_func;
+			bool			datum_typbyval;
+			bool			has_nulls;
+
+			WinGetFrameData(winobj, i, dvalues, isnulls, &numels, &has_nulls);
+
+			datum_typid = function->arg_typid[i];
+			datum_out_func = function->arg_out_func[i];
+			datum_typbyval = function->arg_typbyval[i];
+			PROTECT(el = pg_datum_array_get_r(dvalues, isnulls, numels, has_nulls,
+											  datum_typid, datum_out_func, datum_typbyval));
+
+			/*
+			 * We already set function->nargs arguments
+			 * so we must start with a function->nargs
+			 */
+			SET_VECTOR_ELT(rargs, function->nargs + i, el);
+
+			UNPROTECT(1);
+		}
+
+		/* fnumrows */
+		PROTECT(el = NEW_NUMERIC(1));
+		NUMERIC_DATA(el)[0] = (double) numels;
+		SET_VECTOR_ELT(rargs, m * function->nargs + 0, el);
+		UNPROTECT(1);
+
+		/* prownum */
+		PROTECT(el = NEW_NUMERIC(1));
+		NUMERIC_DATA(el)[0] = (double) WinGetCurrentPosition(winobj) + 1;;
+		SET_VECTOR_ELT(rargs, m * function->nargs + 1, el);
+		UNPROTECT(1);
+	}
+#endif
+
+	UNPROTECT(1);
 	return(rargs);
 }
 
@@ -1663,3 +1803,52 @@ pg_unprotect(int n, char *fn, int ln)
 	unprotect(n);
 }
 #endif /* DEBUGPROTECT */
+
+#ifdef HAVE_WINDOW_FUNCTIONS
+/*
+ * WinGetFrameData
+ *		Evaluate a window function's argument expression on a specified
+ *		window frame, returning an array of Datums for the frame
+ *
+ * argno: argument number to evaluate (counted from 0)
+ * isnull: output argument, receives isnull status of result
+ */
+static void
+WinGetFrameData(WindowObject winobj, int argno, Datum *dvalues, bool *isnulls, int *numels, bool *has_nulls)
+{
+	int64		i = 0;
+
+	*has_nulls = false;
+	for(;;)
+	{
+		Datum	lcl_dvalue;
+		bool	lcl_isnull;
+		bool	isout;
+		bool	set_mark;
+
+		if (i > 0)
+			set_mark = false;
+		else
+			set_mark = true;
+
+		lcl_dvalue = WinGetFuncArgInFrame(winobj, argno, i, WINDOW_SEEK_HEAD,
+										  set_mark, &lcl_isnull, &isout);
+
+		if (!isout)
+		{
+			dvalues[i] = lcl_dvalue;
+			isnulls[i] = lcl_isnull;
+			if (lcl_isnull)
+				*has_nulls = true;
+		}
+		else
+		{
+			*numels = i;
+			break;
+		}
+
+		i++;
+	};
+}
+#endif
+
